import os
import importlib
from salal.core.log import log
from salal.core.config import config

class Handlers:

    #---------------------------------------------------------------------------

    @classmethod
    def load_handlers (cls, directory):
        # Handlers are a way to determine what processing to carry out in a
        # particular instance based on a 'tag' value. To implement a set
        # of handlers, in <directory> have separate .py files for each
        # handler. Each file should create an object called <handler>.
        # The handler object should have a method <get_tags> that returns
        # a list of the tags that should be associated with this particular
        # handler. This method will real all those files, and return a
        # dict where the keys are all the tags, and the values are the
        # corresponding handler objects. Generally each handler object
        # should have one or more additional methods that carry out the
        # actual processing, but what those are and how they are called
        # is up to the code that calls the <load_handler> method.

        handlers = dict()
        with os.scandir(os.path.join(config.system['paths']['salal_root'], directory)) as entries:
            for entry in entries:
                # filter files that start with a period or don't end
                # with .py
                if entry.is_file() and (not entry.name.startswith('.')) and entry.name.endswith('.py') and entry.name != '__init__.py':
                    package_specifier = 'salal.' + os.path.normpath(os.path.join(directory, entry.name)).replace(os.sep, '.').replace('.py', '')
                    log.message('TRACE', 'Loading handler from ' + package_specifier)
                    handler_module = importlib.import_module(package_specifier)
                    for tag in handler_module.handler.get_tags():
                        log.message('TRACE', tag)
                        handlers[tag] = handler_module.handler
        return handlers
    
    #---------------------------------------------------------------------------

handlers = Handlers



import json
import os.path
import glob
import re
import xml.etree.ElementTree as ET
from salal.core.config import config

#------------------------------------------------------------------------------

def load_config(source):
    with open(os.path.join(config.system['paths']['config_root'], source + '.json')) as config_file_fh:
        variables = json.load(config_file_fh)
        return variables

#------------------------------------------------------------------------------

def page_attributes(page_id):
    # we use 'home' instead of a blank id 
    if page_id == 'home':
        page_path = os.path.join(config.system['paths']['content_root'], 'index.xml')
    else:
        matching_pages = glob.glob(config.system['paths']['content_root'] + '/**/' + page_id + '/index.xml', recursive=True)
        if len(matching_pages) > 1:
            raise ValueError('Non-unique page ID "' + page_id + '"')
        elif len(matching_pages) == 0:
            raise ValueError('Can\'t find page with ID "' + page_id + '"')
        page_path = matching_pages[0]
        
    page_root = ET.parse(page_path).getroot()

    # The rest of this is for two auto-generated attributes: 'link' and
    # 'parent'. Note that 'parent' is in the form of a page ID.
    page_root.attrib['link'] = page_path[7:-9]
    page_depth = page_root.attrib['link'].count('/')
    if page_depth > 3:
        parent_path = re.sub(r'[^/]+/\Z', '', page_root.attrib['link'])
        page_root.attrib['parent'] = re.sub(r'\A/.+/([^/]+)/\Z', r'\1', parent_path)
    elif page_depth == 3:
        page_root.attrib['parent'] = re.sub(r'/([^/]+)/[^/]+/\Z', r'\1', page_root.attrib['link'])
    elif page_depth == 2:
        page_root.attrib['parent'] = 'home'
    elif page_depth == 1:
        page_root.attrib['parent'] = None
    return page_root.attrib
    
#------------------------------------------------------------------------------

def emphasize(text):
    return '<em>' + text + '</em>'

#------------------------------------------------------------------------------

def bold(text):
    return '<b>' + text + '</b>'

#------------------------------------------------------------------------------

def inline_link(url, text, destination="external"):
    if destination == "external":
        anchor = '<a href="' + url + '" target="_blank">' + text + '</a>'
    else:
        anchor = '<a href="' + url + '">' + text + '</a>'
    return anchor

#------------------------------------------------------------------------------

def literal_tag(tag):
    return '<![CDATA[&lt;' + tag + '&gt;]]>'

#------------------------------------------------------------------------------

def register_functions(env):
    env.globals['load_config'] = load_config
    env.globals['page_attributes'] = page_attributes
    env.globals['emphasize'] = emphasize
    env.globals['bold'] = bold
    env.globals['inline_link'] = inline_link
    env.globals['literal_tag'] = literal_tag
    
#------------------------------------------------------------------------------



# This recursively expands the templates specified in the XML source file,
# and writes the result to an HTML file. During the expansion, templates
# have access to both the project variables and the attributes and content
# provided with the corresponding tag.
import os.path
import re
import jinja2
from lxml import etree as ET
from salal.core.log import log
from salal.core.config import config
import salal.scripts.file_processing_helpers.custom_jinja_functions as custom_jinja_functions

class XMLHandler:

    #---------------------------------------------------------------------------

    @classmethod
    def get_tags (cls):
        return ['xml']

    #---------------------------------------------------------------------------

    @classmethod
    def configure_modules (cls, node, env):
        module_dirs = [os.path.join(config.system['paths']['design_root'], config.system['paths']['module_dir'])]
        if 'theme_root' in config.system['paths']:
            module_dirs.append(os.path.join(config.system['paths']['theme_root'], config.system['paths']['module_dir']))
        for module in node.attrib['modules'].split():
            # try to locate the module directory
            module_location = None
            for module_dir in module_dirs:
                module_subdir = os.path.join(module_dir, module)
                if os.path.isdir(module_subdir):
                    log.message('TRACE', 'Found module ' + module + ' in ' + module_dir)
                    module_location = module_dir
                    env.loader.searchpath.append(module_subdir)
                    break
            else:
                log.message('ERROR', 'Cannot find module ' + module)

            # if it exists, add module style sheet to the styles list
            for extension, attribute in [('css', 'styles'), ('js', 'scripts')]:
                file_path =  os.path.join(module, module + '.' + extension)
                if os.path.exists(os.path.join(module_location, file_path)):
                    log.message('TRACE', 'Configuring ' + attribute + ' for module ' + module)
                    if attribute not in node.attrib:
                        node.attrib[attribute] = ''
                    else:
                        node.attrib[attribute] += ' '
                    node.attrib[attribute] += os.path.join(os.sep, extension, file_path)

    #---------------------------------------------------------------------------

    @classmethod
    def render_node (cls, node, env, variables):

        # A tage name beginning with _ indicates a preprocessor directive, i.e.,
        # something needs to be done with the node before further processing
        if node.tag.startswith('_'):
            if node.tag == '_include':
                if 'file' in node.attrib:
                    include_tree = ET.parse(node.attrib['file'])
                    include_root = include_tree.getroot()
                    for child in include_root:
                        node.append(child)
                else:
                    raise ValueError('_include directive requires a file attribute')
            else:
                raise ValueError('Unrecognized preprocessor directive ' + node.tag)
        
        # Because of how XML files get parsed, if a node has both
        # children and non-empty text content, we can't guarantee the
        # relative order of them will be preserved in the output. So
        # we check for this and abort if it is encountered.
        if len(node) > 0 and node.text and node.text.strip():
            log.message('ERROR', 'XML nodes with both text and children are not permitted')

        ## Three scenarios for how to set the text content of this node:
    
        # If the node just has text content, expand variables in that
        # content, and use the result as the node's text.
        if node.text and node.text.strip():
            content_template = env.from_string(node.text.strip())
            node.text = content_template.render(variables)
        # If the node has children, call this method recursively on
        # each child. Concatenate the results from those calls, and
        # use that as the node's text.
        elif len(node) > 0:
            node.text = ''
            for child in node:
                node.text += cls.render_node(child, env, variables) + '\n'
        # If the node doesn't have children or non-empty content, set
        # the node's text to the empty string.
        else:
            node.text = ''    

        # Initialize the variables that will be passed to Jinja for rendering
        # the node. We start with whatever variables were passed in, and
        # add a 'this' variable that points to a dict of variables specific
        # to this node. Those include any attributes on the node, as well as
        # a special variable 'this.content' that contains the node text that
        # was set above.
        render_variables = variables.copy()
        render_variables['this'] = { 'content': node.text }
        if node.attrib:
            render_variables['this'].update(node.attrib)

        # Load the template for this node
        if node.tag.startswith('_'):
            template = env.from_string('{{ this.content }}')
        else:
            template = env.get_template(node.tag + '.html')

        # Render the node and return the result
        return template.render(render_variables)
    
    #---------------------------------------------------------------------------

    @classmethod
    def process (cls, tag, source_dir, target_dir, file_stem):

        log.message('TRACE', 'Doing XML expansion')
        # Get the XML source file to be expanded
        xml_root = ET.parse(os.path.join(source_dir, file_stem + '.' + tag)).getroot()
        # We auto-generate a page ID, which is just the name of the directory
        # holding the index.xml file
        page_depth = file_stem.count('/')
        if page_depth == 0:
            page_id = 'home'
        elif page_depth == 1:
            page_id = re.sub(r'/index\Z', '', file_stem)
        else:
            page_id = re.sub(r'\A.+/([^/]+)/index\Z', r'\1', file_stem)
        xml_root.attrib['id'] = page_id
        # Configure the directories to be searched for templates. We
        # add the theme template dir after the local one, so a local
        # template will be found first if there is one
        template_dirs = [os.path.join(config.system['paths']['design_root'], config.system['paths']['template_dir'])] 
        if 'theme_root' in config.system['paths']:
            template_dirs.append(os.path.join(config.system['paths']['theme_root'], config.system['paths']['template_dir']))
        # Initialize Jinja
        env = jinja2.Environment(loader = jinja2.FileSystemLoader(template_dirs), trim_blocks = True, lstrip_blocks = True)
        # If there is a modules attribute on the root, configure modules
        if 'modules' in xml_root.attrib:
            log.message('TRACE', 'Configuring modules for ' + file_stem)
            cls.configure_modules(xml_root, env)
        # Register Salal-specific Jinja functions
        custom_jinja_functions.register_functions(env)
        # Do template expansion on the source file
        xml_root.text = cls.render_node(xml_root, env, config.project)
        # Write the expanded file to the target directory
        with open(os.path.join(target_dir, file_stem + '.html'), mode = 'w', encoding = 'utf-8', newline = '\n') as output_fh:
            output_fh.write(xml_root.text)

    #---------------------------------------------------------------------------

handler = XMLHandler



# This recursively expands the templates specified in the XML source file,
# and writes the result to an HTML file. During the expansion, templates
# have access to both the project variables and the attributes and content
# provided with the corresponding tag.
import os.path
from lxml import etree as ET
from salal.core.log import log
from salal.core.config import config
from salal.core.utilities import utilities

class SVGHandler:

    #---------------------------------------------------------------------------

    @classmethod
    def get_tags (cls):
        return ['svg']

    #---------------------------------------------------------------------------

    @classmethod
    def check_for_def_uses_in_elements(cls, root, attribute, prefix, f_trim, b_trim, def_uses):
        elements = root.findall('.//*[@' + attribute + ']')
        for element in elements:
            if element.attrib[attribute].find(prefix) != -1:
                if b_trim:
                    used_def = element.attrib[attribute][f_trim:b_trim]
                else:
                    used_def = element.attrib[attribute][f_trim:]
                if used_def in def_uses:
                    def_uses[used_def] += 1

    #---------------------------------------------------------------------------

    @classmethod
    def check_for_def_uses_in_styles(cls, root, attribute, f_trim, b_trim, def_uses):
        for element in root.findall('./{*}style[@type="text/css"]'):
            for line in element.text.split('\n'): 
                if line.count(attribute):
                    used_def = line.strip().split()[1][f_trim:b_trim]
                    if used_def in def_uses:
                        def_uses[used_def] += 1

    #---------------------------------------------------------------------------

    @classmethod
    def remove_unused_defs (cls, svg_root):
        def_elements = dict()
        def_uses = dict()
        for element in svg_root.iterfind('./{*}defs/*[@id]'):
            def_elements[element.attrib['id']] = element
            def_uses[element.attrib['id']] = 0
        cls.check_for_def_uses_in_elements(svg_root, 'stroke', 'url(#', 5, -1, def_uses)
        cls.check_for_def_uses_in_elements(svg_root, 'fill', 'url(#', 5, -1, def_uses)
        cls.check_for_def_uses_in_elements(svg_root, 'href', '#', 1, None, def_uses)
        cls.check_for_def_uses_in_styles(svg_root, 'stroke', 5, -2, def_uses)
        cls.check_for_def_uses_in_styles(svg_root, 'fill', 5, -2, def_uses)
        for id in def_uses:
            if def_uses[id] == 0:
                parent = def_elements[id].getparent()
                parent.remove(def_elements[id])

    #---------------------------------------------------------------------------
    # This is a fairly fragile function - it depends on there only being
    # one stylesheet, and it being a specific site-wide css file with a
    # particular name and location on the server. It also uses a hack
    # in lxml to remove the old stylesheet and add a new one.

    @classmethod
    def make_server_ready_stylesheet (cls, svg_root):
        els = svg_root.xpath("preceding-sibling::node()")
        for el in els:
            if hasattr(el, 'target') and (el.target == 'xml-stylesheet'):
                new_el = ET.ProcessingInstruction('xml-stylesheet', 'type="text/css" href="{{ SITE_ROOT }}/css/svg-style.css"')
                svg_root.addprevious(new_el)
                # This weird code is because there is no direct way to remove
                # a processing instruction from the tree. What this code does
                # it moves it from being a sibling of the svg node to a child,
                # at which point it can be removed.
                svg_root.append(el)
                svg_root.remove(el)

    #---------------------------------------------------------------------------

    @classmethod
    def process (cls, tag, source_dir, target_dir, file_stem):

        log.message('TRACE', 'Preparing SVG file')
        tree = ET.parse(os.path.join(source_dir, file_stem + '.' + tag))
        root = tree.getroot()
        cls.remove_unused_defs(root)
        cls.make_server_ready_stylesheet(root)
        tree_string = ET.tostring(tree, pretty_print=True, xml_declaration=True, method="xml").decode()
        output = utilities.substitute_variables(tree_string, config.project)
        with open(os.path.join(target_dir, file_stem + '.' + tag), mode = 'w', encoding = 'utf-8', newline = '\n') as output_fh:
            output_fh.write(output)
            
    #---------------------------------------------------------------------------

handler = SVGHandler



# This handler copies the file from the source to the target directory,
# substituting any references to project variables with their current values.
#
# For advanced users: Technically, the files get the full Jinja
# treatment, so you can put anything in the file that you can put in a
# Jinja template. But the only variables that will be available are
# project variables.
import os.path
import jinja2
from salal.core.log import log
from salal.core.config import config

class SimpleExpansion:
    
    #---------------------------------------------------------------------------

    @classmethod
    def get_tags (cls):
        return ['js', 'css', 'htaccess']
    
    #---------------------------------------------------------------------------

    @classmethod
    def process (cls, tag, source_dir, target_dir, file_stem):
        log.message('TRACE', 'Doing simple expansion')
        env = jinja2.Environment(loader = jinja2.FileSystemLoader(source_dir))
        template = env.get_template(file_stem + '.' + tag)
        output = template.render(config.project)
        with open(os.path.join(target_dir, file_stem + '.' + tag), mode = 'w', encoding = 'utf-8', newline = '\n') as output_fh:
            output_fh.write(output)

    #---------------------------------------------------------------------------

handler = SimpleExpansion



import os
from salal.core.log import log
from salal.core.config import config
from salal.core.utilities import utilities 
from salal.core.file_processing import file_processing

class Build:
    
    #---------------------------------------------------------------------------

    @classmethod
    def get_tags (cls):
        return ['build']
    
    #---------------------------------------------------------------------------

    @classmethod
    def process_content (cls):
        for file_type in config.system['content_file_types']:
            log.message('TRACE', 'Looking for ' + file_type + 'files')
            for file_relative_path in utilities.find_files_by_extension(config.system['paths']['content_root'], file_type):
                # create the target directory if it doesn't exist
                os.makedirs(os.path.join(config.system['paths']['profile_build_dir'], os.path.dirname(file_relative_path)), exist_ok = True)
                log.message('INFO', os.path.join(config.system['paths']['content_root'], file_relative_path))
                file_processing.process(config.system['paths']['content_root'], config.system['paths']['profile_build_dir'], file_relative_path)

    #---------------------------------------------------------------------------

    @classmethod
    def process_resources (cls):
        # Copy all the files in the resources directory to the build
        # directory, processing them according to the appropriate
        # file processing handler.
        #
        # We write files to the same relative path in the build
        # directory as they had in the resources directory. So, for
        # profile <test>, /resources/js/app.js will become
        # /build/test/js/app.js.
        #
        # Theme files get processed first, so they can be overridden
        # by local files. This is accomplished simply by overwriting
        # the theme version of the file.
        resource_dirs = [os.path.join(config.system['paths']['design_root'], config.system['paths']['resource_dir'])]
        if 'theme_root' in config.system['paths']:
            resource_dirs.insert(0, os.path.join(config.system['paths']['theme_root'], config.system['paths']['resource_dir']))
        for resource_dir in resource_dirs:
            # Check if the resource directory exists before processing it
            if not os.path.isdir(resource_dir):
                continue
            log.message('DEBUG', 'Processing resources from ' + resource_dir)
            resource_files = utilities.find_files(resource_dir)
            for file_relative_path in resource_files:
                # create the target directory if it doesn't exist
                os.makedirs(os.path.join(config.system['paths']['profile_build_dir'], os.path.dirname(file_relative_path)), exist_ok = True)
                log.message('INFO', os.path.join(resource_dir, file_relative_path))
                file_processing.process(resource_dir, config.system['paths']['profile_build_dir'], file_relative_path)

    #---------------------------------------------------------------------------

    @classmethod
    def process_modules (cls):
        # Copy module files to the build directory, processing them
        # according to the appropriate file processing handler.
        #
        # The destination directory that is used depends on the system
        # variable <module_mappings>, which is used to set an appropriate
        # destination based on module name and/or file extension.
        #
        # Theme files get processed first, so they can be overridden
        # by local files. This is accomplished simply by overwriting
        # the theme version of the file.
        file_types_to_copy = {'css': 'css', 'js': 'js', 'py': 'py'}
        module_dirs = [os.path.join(config.system['paths']['design_root'], config.system['paths']['module_dir'])]
        if 'theme_root' in config.system['paths']:
            module_dirs.insert(0, os.path.join(config.system['paths']['theme_root'], config.system['paths']['module_dir']))
        for module_dir in module_dirs:
            # Check if the module directory exists before processing it
            if not os.path.isdir(module_dir):
                continue
            log.message('DEBUG', 'Processing modules from ' + module_dir)
            module_subdirs = utilities.find_subdirectories(module_dir)
            for module_subdir in module_subdirs:
                log.message('TRACE', 'Found module ' + module_subdir)
                for file_type in file_types_to_copy:
                    log.message('TRACE', 'Looking for ' + file_type + 'files in ' + module_subdir)
                    source_dir = os.path.join(module_dir, module_subdir)
                    target_dir = os.path.join(config.system['paths']['profile_build_dir'], file_type, module_subdir)
                    module_files = utilities.find_files_by_extension(source_dir, file_type)
                    for module_file in module_files:
                        log.message('INFO', os.path.join(source_dir, module_file))
                        # create the target directory if it doesn't exist
                        os.makedirs(target_dir, exist_ok = True)
                        file_processing.process(source_dir, target_dir, module_file)
        
    #---------------------------------------------------------------------------

    @classmethod
    def execute (cls, tag):
        
        cls.process_content()
        cls.process_resources()
        cls.process_modules() 

    #---------------------------------------------------------------------------

handler = Build
